/* proxycheck.c: open proxy checker, main program.
 * Michael Tokarev  <mjt@corpit.ru>.
 * This code may be freely used and distributed according to
 * the terms of General Public License (GPL) version 2 or later.
 */

#ifdef _AIX
#pragma alloca
#endif

#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include "event.h"
#include "pxy.h"

#ifndef INADDR_NONE
# define INADDR_NONE ((in_addr_t)0xffffffff)
#endif

typedef struct {
  const pxyproto_t *proto;	/* proxy protocol to use */
  ipport_t port;		/* port to connect to */
} pxyprotoport_t;

typedef struct pxyhost {
  struct in_addr addr;		/* IP address of proxy */
  pxyprotoport_t *pps;		/* array of proto:ports to test */
  pxyprotoport_t *cpp;		/* next proto:port to test */
  int nopen;			/* numbef of open ports found */
  int nactive;			/* number of active connections */
  struct pxyhost *next;		/* next host to test */
} pxyhost_t;

static char *progname;
static int verbose;		/* verbosity level */
static int maxconn;		/* max number of connections */
struct sockaddr_in baddr;	/* bind address */
static pxyhost_t *hosts;	/* host list */
static int nhosts;		/* number of hosts to check */
static pxyhost_t *chost;	/* current host */
static pxyprotoport_t *pps;	/* default array of protos:ports */
static int ntodo;		/* number of hosts and total ports */
static int nactive;		/* number of active connections */
static int nopen;		/* number of open proxies found */
static long nread;		/* number of bytes read */
static ipaddr_t dst;		/* destination address */

typedef struct {
  char *name;
  void (*connh)(int fd, ev_t e, pxyconn_t *c);
  int (*init)(char *arg);
  char *arg;
  char *descr;
} pxycheck_t;
extern const pxycheck_t checks[];
static const pxycheck_t *check;

static void
err(int errnum, const char *fmt, ...) {
  char buf[1024];
  int l = sprintf(buf, "%s: ", progname);
  va_list ap;
  va_start(ap, fmt);
  l += vsprintf(buf + l, fmt, ap);
  va_end(ap);
  if (errnum)
    l += sprintf(buf + l, ": %s\n", strerror(errno));
  else
    buf[l++] = '\n';
  write(2, buf, l);
  exit(1);
}

void
pxyinfo(const pxyconn_t *c, int level, const char *fmt, ...) {
  if (level <= verbose) {
    char buf[512];
    int len;
    va_list ap;
    if (c)
      len = sprintf(buf, "%s:%s:%d: ",
		    inet_ntoa(c->pxy.addr), c->proto->name, c->pxy.port);
    else
      len = 0;
    va_start(ap, fmt);
    len += vsprintf(buf + len, fmt, ap);
    va_end(ap);
    buf[len++] = '\n';
    write(2, buf, len);
  }
}

void pxyvio(pxyconn_t *c, int level, int direction,
	    const char *s, int len) {
  if (level <= verbose) {
    char buf[90];
    char *bs, *bp;
    const char *e = s + len;
    char *const be = buf + sizeof(buf) - 1 - 4;
    bs = buf + sprintf(buf, "%s:%s:%d: %s ",
		       inet_ntoa(c->pxy.addr), c->proto->name, c->pxy.port,
		       direction > 0 ? ">>" : "<<");
    bp = bs;
#define flushb() \
   do { *bp++ = '\n'; write(2, buf, bp - buf); bp = bs; } while(0)
    while(s < e) {
      if (bp >= be)
	flushb();
      switch(*s) {
      case '\n': *bp++ = '\\'; *bp++ = 'n'; break;
      case '\r': *bp++ = '\\'; *bp++ = 'r'; break;
      case '\t': *bp++ = '\\'; *bp++ = 't'; break;
      case '\\': *bp++ = '\\'; *bp++ = '\\'; break;
      case 0: *bp++ = '\\'; *bp++ = '0'; break;
      default:
	if (*s < ' ' || *s >= 0177)
	  bp += sprintf(bp, "\\0%o", (unsigned char)*s);
	else
	  *bp++ = *s;
      }
      if (*s == '\n')
	flushb();
      ++s;
    }
    if (bp > bs)
      flushb();
  }
#undef flushb
}

void *emalloc(unsigned size) {
  void *ptr = malloc(size);
  if (!ptr)
    err(0, "out of memory (%d bytes)", size);
  return ptr;
}

void *erealloc(void *ptr, unsigned size) {
  ptr = realloc(ptr, size);
  if (!ptr)
    err(0, "out of memory (%d bytes)", size);
  return ptr;
}

static int
satoi(const char *s) {
  int c = 0;
  if (*s < '0' || *s > '9') return -1;
  do
    c = c * 10 + (*s++ - '0');
  while (*s >= '0' && *s <= '9');
  return *s ? -1 : c;
}

static in_addr_t
hostaddr(const char *s) {
  in_addr_t a = inet_addr(s);
  if (a == INADDR_NONE) {
    struct hostent *he = gethostbyname(s);
    if (!he) return 0;
    if (he->h_addrtype != AF_INET ||
	he->h_length != 4)
      return 0;
    memcpy(&a, he->h_addr_list[0], 4);
  }
  return a;
}

static ipport_t
portnum(const char *s) {
  int p = satoi(s);
  return p < 1 || p > 0xffff ? 0 : (ipport_t)p;
}

static pxyhost_t *
findhost(register in_addr_t addr) {
  register pxyhost_t *h;
  for (h = hosts; h; h = h->next)
    if (h->addr.s_addr == addr)
      return h;
  h = (pxyhost_t*) emalloc(sizeof(*h));
  memset(h, 0, sizeof(*h));
  h->addr.s_addr = addr;
  h->next = hosts;
  hosts = h;
  return h;
}

static void
donehost(pxyhost_t *h) {
  pxyhost_t *hp;
  pxyinfo(NULL, 4, "%s: done, numopen=%d", inet_ntoa(h->addr), h->nopen);
  if (h->pps != pps)
    free(h->pps);
  if (h == hosts)
    hosts = h->next;
  else {
    hp = hosts;
    while(hp->next != h)
      hp = hp->next;
    hp->next = h->next;
  }
  hp = h->next;
  if (chost == h)
    chost = hp ? hp : hosts;
  free(h);
  --nhosts;
}

static void
usage(const char *fmt, ...) {
  char buf[256];
  int l = sprintf(buf, "%s: ", progname);
  va_list ap;
  va_start(ap, fmt);
  l += vsprintf(buf + l, fmt, ap);
  va_end(ap);
  l += sprintf(buf + l, "\n%s: `%s -h' for help\n", progname, progname);
  write(2, buf, l);
  exit(1);
}

/* parse list of proto/ports */

static ipport_t
nextport(char *s) {
  s = strtok(s, ",|; \t");
  if (s) {
    ipport_t p = portnum(s);
    if (!p)
      usage("invalid port `%s'", s);
    return p;
  }
  else
    return 0;
}

static pxyprotoport_t *
addprotoport(pxyprotoport_t *pp, ipport_t port, const pxyproto_t *proto) {
  pxyprotoport_t *p;
  if (!pp) {
    pp = (pxyprotoport_t*)emalloc(sizeof(*pp) * 5);
    p = pp;
  }
  else {
    int n;
    for (n = 0; pp[n].port; ++n)
      if (pp[n].port == port && pp[n].proto == proto)
	return pp;
    if (!(n & 3))
      pp = (pxyprotoport_t*)erealloc(pp, sizeof(*pp) * (n + 5));
    p = pp + n;
  }
  p->port = port;
  p->proto = proto;
  p[1].port = 0;
  p[1].proto = NULL;
  return pp;
}

static pxyprotoport_t *
parseprotoport(pxyprotoport_t *pp, char *s) {
  char *p;
  const pxyproto_t *proto;
  const ipport_t *portp;
  ipport_t port;
  s = strcpy((char*)alloca(strlen(s)+1), s);
  if ((p = strchr(s, ':')) != NULL) {
    *p++ = '\0';
    if ((proto = pxyfindproto(s)) == NULL)
      usage("invalid protocol `%s'", s);
    if (!(port = nextport(p))) {
      for (portp = proto->defports; *portp; ++portp)
	pp = addprotoport(pp, *portp, proto);
    }
    else {
      do pp = addprotoport(pp, port, proto);
      while((port = nextport(NULL)) != 0);
    }
  }
  else { /* no protocol specified */
    if (!(port = nextport(s)))
      usage("invalid protocol/port specification");
    do {
      int found = 0;
      for (proto = pxyprotos; proto->name; ++proto)
	for (portp = proto->defports; *portp; ++portp)
	  if (*portp == port) {
	    pp = addprotoport(pp, port, proto);
	    found = 1;
	    break;
	  }
      if (!found) {
	for (proto = pxyprotos; proto->name; ++proto)
	  pp = addprotoport(pp, port, proto);
      }
    } while((port = nextport(NULL)) != 0);
  }
  return pp;
}

static int
numprotoports(const pxyprotoport_t *pp) {
  const pxyprotoport_t *ppp = pp;
  while(ppp->port) ++ppp;
  return ppp - pp;
}

static void
init(int argc, char **argv) {
  int c;
  int timeout = 0;
  pxyhost_t *h;
  const pxyproto_t *proto;
  const ipport_t *portp;
  pxyprotoport_t *pp;
  char *p;
  int npps;
  int readin = 0;
  char *check_arg = NULL;

  if ((progname = strrchr(argv[0], '/')) != NULL)
    argv[0] = ++progname;
  else
    progname = argv[0];

  while((c = getopt(argc, argv, "vd:c:p:b:t:m:i:h")) != EOF)

    switch(c) {

    case 'v':
      ++verbose;
      break;

    case 'p':
      pps = parseprotoport(pps, optarg);
      break;

    case 'd':
      if (!(p = strchr(optarg, ':')))
	usage("destination port missing in `%s'", optarg);
      *p = '\0';
      if (!(dst.addr.s_addr = hostaddr(optarg)))
	usage("unknown destination host `%s'", optarg);
      if (!(dst.port = portnum(p+1)))
	usage("invalid destination port `%s'", p+1);
      *p = ':';
      break;

    case 'c':
      if ((p = strchr(optarg, ':')) != NULL)
	*p = '\0';
      for(check = checks; ; ++check)
	if (!check->name)
	  usage("unknown check `%s'", optarg);
	else if (strcmp(optarg, check->name) == 0)
	  break;
      if (p)
	check_arg = p + 1, *p = ':';
      else
	check_arg = NULL;
      break;

    case 'b':
      if (!(baddr.sin_addr.s_addr = hostaddr(optarg)))
	usage("unknown host `%s'", optarg);
      baddr.sin_family = AF_INET;
      break;

    case 't':
      if ((timeout = satoi(optarg)) < 1 || timeout > 1000)
	usage("invalid timeout `%s'", optarg);
      break;

    case 'm':
      if ((maxconn = satoi(optarg)) < 1)
	usage("invalid maximum number of connections `%s'", optarg);
      break;

    case 'i':
      readin = 1;
      break;

    case 'h':
      printf(
"%s: Open proxy checker version " VERSION_STR "\n"
"Usage is: `%s options host[:proto_port_spec]...'\n"
"where options are:\n"
" -h - print this help and exit\n"
" -d dsthost:dstport - destination to connect to (required)\n"
" -c check[:params] - method to check proxy (required, see below)\n"
" -p proto_port_spec - proxy port/protocol specification\n"
"   (may be used more than once) - see below\n"
" -t timeout - general timeout in secounds, default %d\n"
" -m maxconn - maximum number of parallel connections\n"
" -b bindaddr - bind to specified address\n"
"\n"
"proto_port_spec is in the form [proto:][port,port,...].\n"
"If portlist is omitted, default ports for given protocols\n"
"will be tried; if proto is omitted, either all protocols will\n"
"be tried (if port is not known), or the protocols which are\n"
"assotiated with this port.\n"
"\n"
"The following proto:ports are recognized and tried by default:\n"
, progname, progname, pxytimeout / 1000);
      for(proto = pxyprotos; proto->name; ++proto) {
	printf(" %s(%s): ", proto->name, proto->aname);
	for(portp = proto->defports;; ) {
	  printf("%d", *portp++);
	  putchar(*portp ? ',' : '\n');
	  if (!*portp) break;
	}
      }
      printf("\nThe following checks are available:\n");
      for(check = checks; check->name; ++check)
	printf(" %s%s - %s\n", check->name,
	       check->arg ? check->arg : "",
	       check->descr);
      exit(0);

    default:
      err(0, "`%s -h' for help", progname);
    }

  argc -= optind; argv += optind;
  if (!*argv && !readin)
    usage("no host(s) to check specified");

  if (!dst.port)
    usage("no destination (-d) specified");
  if (!check)
    usage("no action (-c) specified");

  while(*argv) {
    char *n = *argv++;
    in_addr_t addr;
    if ((p = strchr(n, ':'))) *p = '\0';
    if (!(addr = hostaddr(n)))
      usage("invalid IP address `%s'", n);
    h = findhost(addr);
    if (p) {
      *p++ = ':';
      if (*p)
	h->pps = parseprotoport(h->pps, p);
      else
	h->nopen = 1; /* remember to add std pps */
    }
    else
      h->nopen = 1; /* remember to add std pps */
  }

  check->init(check_arg);

  if (!pps) { /* no explicit ports given, pick up defaults */
    npps = 0;
    for (proto = pxyprotos; proto->name; ++proto)
      for(portp = proto->defports; *portp; ++portp)
	++npps;
    pp = pps = (pxyprotoport_t *)emalloc(sizeof(*pp) * (npps + 1));
    for (proto = pxyprotos; proto->name; ++proto)
      for(portp = proto->defports; *portp; ++portp) {
	pp->proto = proto;
	pp->port = *portp;
	++pp;
      }
    pp->port = 0;
    pp->proto = NULL;
  }
  else
    npps = numprotoports(pps);

  for (h = hosts; h; h = h->next) {
    ++nhosts;
    if (!h->pps)
      h->pps = pps, ntodo += npps;
    else {
      if (h->nopen) {
	for (pp = pps; pp->port; ++pp)
	  h->pps = addprotoport(h->pps, pp->port, pp->proto);
      }
      ntodo += numprotoports(h->pps);
    }
    h->nopen = 0;
    h->cpp = h->pps;
#if 0
    if (verbose) {
      pxyprotoport_t *pp;
      printf("%s:", inet_ntoa(h->addr));
      for(pp = h->pps; pp->port; ++pp)
	printf(" %s:%d", pp->proto->name, pp->port);
      putchar('\n');
    }
#endif
  }
  pxyinfo(NULL, 1,
	  "To check: hosts=%d, proto:ports=%d, host:proto:ports=%d",
	  nhosts, npps, ntodo);

  if (timeout)
    pxytimeout = timeout * 1000;
  chost = hosts;

  signal(SIGPIPE, SIG_IGN);
}

static int
nextconn(void) {
  pxyconn_t *c;
  pxyprotoport_t *pp;
  pxyhost_t *h;
  int fd;
  struct sockaddr_in sin;

  if (!ntodo)
    return 0;

  if (maxconn && nactive >= maxconn)
    return 0;

  fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (fd < 0 || fd >= FD_SETSIZE) {
    if (fd < 0 && errno != EMFILE)
      err(errno, "unable to create socket");
    if (!maxconn || maxconn > nactive) {
      pxyinfo(NULL, 1, "limiting max number of connections to %d",
	      nactive);
      maxconn = nactive;
    }
    return 0;
  }

  while(!chost->cpp->port)
    if (!(chost = chost->next))
      chost = hosts;

  pp = chost->cpp++;
  h = chost;
  if (!(chost = chost->next))	/* move to next host to distribute load */
    chost = hosts;
  --ntodo;

  if (baddr.sin_addr.s_addr &&
      bind(fd, (struct sockaddr*)&baddr, sizeof(baddr)) < 0)
    err(errno, "unable to bind to %s", inet_ntoa(baddr.sin_addr));
  if (fcntl(fd, F_SETFL, O_NONBLOCK|O_RDWR) < 0)
    /*fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)*/
    err(errno, "unable to set socket flags");

  c = pxynew();
  if (!c)
    abort();

  c->pxy.addr = h->addr;
  c->pxy.port = pp->port;
  c->proto = pp->proto;
  c->dst = dst;
  c->data = h;

  ++h->nactive; ++nactive;

  /* from now on, we should expect to be called recursively:
   * do not touch chost and pp anymore */

  if (pxystart(fd, c) != 0)
    err(errno, "unable to enqueue testing request");

  sin.sin_family = AF_INET;
  sin.sin_addr = h->addr;
  sin.sin_port = htons(pp->port);
  if (connect(fd, (struct sockaddr*)&sin, sizeof(sin)) != 0 &&
      errno != EINPROGRESS)
    pxycerr(fd, EVENT_ERROR, c);

  return 1;
}

static void
freeconn(int fd, pxyconn_t *c) {
  pxyhost_t *h = (pxyhost_t *)c->data;
  nread += c->nread;
  --h->nactive;
  --nactive;
  pxyinfo(c, 5, "done, nactive=%d ntodo=%d", nactive, ntodo);
  pxyfree(fd, c);
  close(fd);
  if (!h->nactive && !h->cpp->port)
    donehost(h);
  nextconn();
}

static void
isopen(int fd, pxyconn_t *c, int is_open) {
  pxyhost_t *h = (pxyhost_t *)c->data;
  printf("%s %s:%d %s\n",
	 inet_ntoa(c->pxy.addr), c->proto->name, c->pxy.port,
	 is_open ? "open" : "closed");
  if (is_open && !(h->nopen++))
    ++nopen;
  freeconn(fd, c);
}

void pxyaction(int fd, pxyconn_t *c, int result) {
  if (!result) {
    c->state = 1;
    pxyrenew(fd, c, 0);
    ev_reqio(fd, EVENT_READ, (void(*)(int,ev_t,void*))check->connh, c);
    check->connh(fd, EVENT_READ, c);
  }
  else {
    if (result < 0)
      pxyinfo(c, c->state ? 3 : 4, c->state ? "%s" : "connect: %s",
	      errno ? strerror(errno) : "EOF");
    else if (result > 1) {
      isopen(fd, c, 0);
      return;
    }
    /*else
      pxyinfo(c, 3, "seems not to be open");*/
    freeconn(fd, c);
  }
}

int main(int argc, char **argv) {
  time_t start;

  init(argc, argv);

  start = time(NULL);

  while(nextconn())
    ;

  while(ev_hasevents())
    if (ev_wait(-1) < 0 && errno != EINTR)
      err(errno, "ev_wait");

  pxyinfo(NULL, 1, "NumOpen=%d NRead=%ld Time=%d",
	  nopen, nread, time(NULL) - start);

  return nopen ? 100 : 0;
}

static char *sendstr;		/* send this string... */
static char *expstr;		/* and expect this */
static unsigned explen;		/* length of expstr */

static void
expecth(int fd, ev_t e, pxyconn_t *c) {
  int l;
  if (pxycerr(fd, e, c))
    return;
  if (c->state == 1) {
    int l;
    if (sendstr)
      l = sprintf(pxybuf, "%s\r\n", sendstr);
    else
      l = sprintf(pxybuf, "%s:%s:%d\r\n",
		  c->proto->name, inet_ntoa(c->pxy.addr), c->pxy.port);
    c->state = 2;
    pxywrite(fd, c, pxybuf, l, 3);
  }
  else if ((l = pxyreadnext(fd, c, 3)) <= 0)
    ;
  else if (memmem(pxybuf, l, expstr, explen))
    isopen(fd, c, 1);
  else if (!c->proto->check || !c->proto->check(fd, c, pxybuf, l))
    pxysave(fd, c, pxybuf, l, explen);
}

static int expecti(char *arg) {
  char *p = strchr(arg, ':');
  if (!p || !p[1])
    usage("send:expect strings expected");
  if (p != arg) {
    sendstr = arg;
    *p++ = '\0';
    arg = p;
  }
  else
    ++arg;
  if (!(expstr = arg) || !(explen = strlen(arg)))
    usage("specify a string to expect");
  else if (explen >= sizeof(((pxybuf_t*)0)->buf))
    usage("expect string is too long");
  return 0;
}

#define DSBL_COOKIE_PORT 200
#define DSBL_COOKIE_LEN 32
static char dsblcookie[DSBL_COOKIE_LEN+1];
static char *dsbluser, *dsblpass, *dsblrcpt, *dsblfrom;

static void
dsblh(int fd, ev_t e, pxyconn_t *c) {
  /* state:
   * 1 - connected, send helo
   * 2 - wait for initial 2xx reply, send from/to/data commands
   * 3 - wait for 3xx reply to data, send message (renew timer)
   * 4 - wait for 2xx reply
   */
  int l;
  char *p;
  if (pxycerr(fd, e, c)) return;
  if (c->state == 1) {
    l = sprintf(pxybuf, "HELO [%s]\r\n", inet_ntoa(c->pxy.addr));
    c->state = 2;
    pxywrite(fd, c, pxybuf, l, 3);
  }
  else if (c->state == 2) {
    if ((l = pxyreadnext(fd, c, 3)) <= 0)
      ;
    else if ((p = (char*)memmem(pxybuf, l, "\r\n2", 3)) /*&& p[5] == ' '*/) {
      l = sprintf(pxybuf,
		  "MAIL FROM:<%s>\r\n"
		  "RCPT TO:<%s>\r\n"
		  "DATA\r\n",
		  dsblfrom, dsblrcpt);
      c->state = 3;
      pxywrite(fd, c, pxybuf, l, 3);
    }
    else if (!c->proto->check || !c->proto->check(fd, c, pxybuf, l))
      pxysave(fd, c, pxybuf, l, 10);
  }
  else if (c->state == 3) {
    if ((l = pxyreadnext(fd, c, 3)) <= 0)
      ;
    else if ((p = (char*)memmem(pxybuf, l, "\r\n3", 3)) /*&& p[3] == ' '*/) {
      char pxyaddr[sizeof("255.255.255.255")];
      strcpy(pxyaddr, inet_ntoa(c->pxy.addr));
      l = sprintf(pxybuf,
"Message-ID: <%s@%s>\r\n"
"To: <%s>\r\n"
"Subject: Open %s Proxy test message\r\n"
"\r\n"
"DSBL LISTME: %s %s\r\n"
"%s\r\n"
"Port %d, connect to %s\r\n"
"DSBL END\r\n"
"\r\n"
".\r\n",
              dsblcookie, dsblfrom, /* Message-ID */
              dsblrcpt, /* To: */
              c->proto->name, /* Subj */
              c->proto->name, pxyaddr, /* dsbl listme */
              dsblcookie, /* dsbl cookie */
              c->pxy.port, inet_ntoa(c->dst.addr));
      if (pxywrite(fd, c, pxybuf, l, 6) == 0) {
	pxyinfo(c, 2, "sending message");
	c->state = 4;
	if (pxysave(fd, c, NULL, 0, 0) == 0)
	  pxyrenew(fd, c, 0);
      }
    }
    else
      pxysave(fd, c, pxybuf, l, 10);
  }
  else {
    if ((l = pxyreadnext(fd, c, 3)) <= 0)
      ;
    else if (pxybuf[0] == '2' && pxybuf[3] == ' ')
      isopen(fd, c, 1);
    else
      pxysave(fd, c, pxybuf, l, 10);
  }
}

static char *egetenv(const char *name, char *def) {
  char *v = getenv(name);
  return v ? v : def;
}

static int
dsbli(char *arg) {
  if (arg && *arg)
    usage("no argument for dsbl expected");

  dsbluser = egetenv("DSBL_USER", "anonimous");
  dsblpass = egetenv("DSBL_PASS", "");
  dsblfrom = egetenv("DSBL_FROM", dsbluser);
  dsblrcpt = egetenv("DSBL_RCPT", "listme@listme.dsbl.org");
  if (!(arg = getenv("DSBL_COOKIE"))) {
    char *chost = egetenv("DSBL_COOKIE_HOST", "cookie.dsbl.org");
    struct sockaddr_in sin;
    int fd;
    int l;

    sin.sin_family = AF_INET;
    sin.sin_port = htons(DSBL_COOKIE_PORT);
    if (!(sin.sin_addr.s_addr = hostaddr(chost)))
      err(0, "unknown DSBL cookie host %s", chost);
    if ((fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0 ||
	connect(fd, (struct sockaddr*)&sin, sizeof(sin)) < 0)
      err(errno, "unable to connect to cookie server");
    l = sprintf(pxybuf, "%s\n%s\n", dsbluser, dsblpass);
    if (send(fd, pxybuf, l, 0) < 0 ||
	recv(fd, dsblcookie, DSBL_COOKIE_LEN, 0) != DSBL_COOKIE_LEN)
      err(errno, "unable to obtain cookie");
    close(fd);
    dsblcookie[DSBL_COOKIE_LEN] = 0;
  }
  else if (strlen(arg) != DSBL_COOKIE_LEN)
    usage("invalid dsbl cookie `%s'", arg);
  else
    strcpy(dsblcookie, arg);
  pxyinfo(NULL, 4, "dsbl cookie: %s", dsblcookie);
  return 0;
}

const pxycheck_t checks[] = {
  { "chat", expecth, expecti, ":sendstr:expectstr",
    "perform a little chat: send `sendstr'\n"
    "   to the remote system  and assume proxy is open is `expectstr'\n"
    "   is returned.  `sendstr' defaults to proto:ip:port"
  },
  { "dsbl", dsblh, dsbli, NULL,
    "attempt to submit proxy to DSBL-like system\n"
    "   DSBL settings are expected to be in environment:\n"
    "\t$DSBL_USER - username (anonimous)\n"
    "\t$DSBL_PASS - password (default is empty)\n"
    "\t$DSBL_COOKIE_HOST - cookie server (cookie.dsbl.org)\n"
    "\t$DSBL_COOKIE - already obtained DSBL cookie\n"
    "\t$DSBL_RCPT - recipient (listme@listme.dsbl.org)\n"
    "\t$DSBL_FROM - sender address (nobody)"
  },
  {0,0,0,0,0}
};
